package edu.udo.cs.ie.cowbots.bdi.brf.operators;

import java.util.*;

import edu.udo.cs.ie.cowbots.bdi.brf.*;
import edu.udo.cs.ie.cowbots.logic.*;
import edu.udo.cs.ie.cowbots.logic.solver.*;

/**
 * implementation of update programs based
 * on causal rejection (proposed by eiter et. al).
 *  
 * @author Thomas
 *
 */
public class CausalRejection implements IInferenceOperator {

	protected ELP updProgram = null;
	
	@Override
	public ELP createBeliefState(List<InformationObject> ioList) {
		
		int counter = 1;
		
		ELP updSeq = new ELP();
		
		Set<ELPLiteral> headLits = new HashSet<ELPLiteral>();
		
		// loop over the update sequence
		Iterator<InformationObject> ioIter = ioList.iterator();
		while( ioIter.hasNext() ) {
			
			InformationObject io = ioIter.next();
			boolean lastIO = !ioIter.hasNext();
				
			// inspect program
			ELP update = io.getELP();
			Iterator<ELPRule> ruleIter = update.getRules().iterator();
			while( ruleIter.hasNext() ) {	
				ELPRule r = ruleIter.next();
				
				boolean lastUpdate = !ioIter.hasNext() && !ruleIter.hasNext();
				
				// simply add constraint
				if (r.isConstraint()) {					
					updSeq.addRule(r);
					continue;
				}
			
				// store head literals
				for (int i = 0; i < r.nHead(); i++)
					headLits.add(r.getHead(i));
				
				
				// rewrite rule:
				// a :- B   =>  a_indexed(i) :- B, rej(i) 
				ELPAtom rejLiteral = new ELPAtom("rej",""+counter);
				ELPRule bodyRule = new ELPRule();
				//bodyRule.addHead( indexLiteral( head, counter ) );
				for (int i = 0; i < r.nHead(); i++) 
					bodyRule.addHead( indexLiteral( r.getHead(i),counter));
				bodyRule.addBody(r.getBody());
				if (!lastUpdate)
					bodyRule.addBody( new NotLiteral( rejLiteral ));
				updSeq.addRule(bodyRule);
				 
				// reject rule
				// rej :- B(r), ~H(r)
				ELPRule rejRule = new ELPRule();
				rejRule.addHead(rejLiteral);
				rejRule.addBody(r.getBody());
				for (int i = 0; i < r.nHead(); i++)
					rejRule.addBody( indexLiteral( inverse(r.getHead(i)), counter+1 ));
				//rejRule.addBody( indexLiteral( inverse(head), counter+1 ));
				updSeq.addRule(rejRule);
			
				++counter;
			}
			
			
		}
		
		// link literals
		for (ELPLiteral l : headLits) {
			ELPRule r = new ELPRule();
			r.addHead(l);
			r.addBody(indexLiteral(l,1));
			updSeq.addRule(r);
			for (int i = 1; i < counter-1; i++) {
				r = new ELPRule();
				r.addHead( indexLiteral(l,i) );
				r.addBody( indexLiteral(l,i+1) );
				updSeq.addRule(r);
			}
		}
		
		updProgram = updSeq;
		
		return updSeq;
	}
	
	/**
	 * creates an indexed version of a literal, with proper sign
	 * 
	 * @param l			literal a(X) to index
	 * @param index		index of literal
	 * @return			a_index_p(X,index) or a_index_n(X,index)
	 * 
	 */
	protected ELPLiteral indexLiteral( ELPLiteral l, int index ) {
		ELPAtom la = l.getAtom();
		String sign = l.isStrictNegated()?"n":"p";
		ELPAtom a = new ELPAtom(la.getPredicate()+"_index_"+sign, la.getArity()+1 );
		for (int i = 0; i < la.getArity(); i++)
			a.setTerm(i, la.getTerm(i));
		a.setTerm(la.getArity(), ""+index);
		
		return a;
	}
	
	/**
	 * returns the negation of a positive or strict negated literal 
	 * @param l
	 * @return
	 */
	protected ELPLiteral inverse( ELPLiteral l ) {
		if (l.isStrictNegated())
			return l.getAtom();
		else
			return new NegLiteral(l.getAtom());
	}
	

	@Override
	public AnswerSet selectAnswerSet(List<AnswerSet> asl, DLV dlv) {
		// check if we have something to test for
		ELP testProgram = buildTestProgram(this.updProgram);
		if (testProgram == null)
			return asl.get(0);
		
		List<AnswerSet> minimalAS = new LinkedList<AnswerSet>();
		
		// test every answer set
		for (AnswerSet as : asl) {
			ELP tp = new ELP();
			
			// iterate over answer set, fetch reject literals
			for (ELPLiteral l : as.literals) {
				
				if (!l.getAtom().getPredicate().equals("rej"))
					continue;
				
				ELPRule r = new ELPRule();
				int i = l.getAtom().getTermAsInt(0);
				r.addHead(new ELPAtom("ok", i));
				r.addBody(new NotLiteral (l));
				r.addBody(new ELPAtom("eq", ""+(i+1)));
				
				tp.addRule(r);
				
			}
			
			tp.add(testProgram);
			
			// no model => strictly minimal as
			List<AnswerSet> models = dlv.computeModels(tp, 5);
			if (models.size() == 0) {
				minimalAS.add(as);
			}
		}
		
		// any as is good at this point
		if (minimalAS.size() > 0)
			return minimalAS.get(0);
		
		return null;
	}
	

	/**
	 * this method creates a test program for
	 * strictly minimal answer sets
	 * 
	 * @param up update sequence to create test for
	 * @return test program
	 */
	protected ELP buildTestProgram(ELP up) {
		if (up == null)
			return null;
		
		ELP ret = new ELP();
		
		int i = 0;
		int imax = 0;
		for (ELPRule r : up.getRules()) {
			ELPAtom rej = isRejRule(r);
			if (rej == null)
				continue;
			
			// create constraint for minimality test
			ELPRule cr = new ELPRule();
			cr.addBody(rej);
			cr.addBody( new NotLiteral( new ELPAtom("s", rej.getTerms())));
			i = rej.getTermAsInt(0);
			imax = (i>imax)?i:imax;
			cr.addBody( new ELPAtom("eq", ""+(i+1)));
			ret.addRule(cr);
		}
		
		for (int j = 1; j <= imax; j++) {
			ELPRule r = new ELPRule();
			r.addHead(new ELPAtom("eq",""+j));
			r.addBody(new ELPAtom("eq",""+(j+1)) );
			r.addBody(new NotLiteral( new ELPAtom("ok",""+j)));
			ret.addRule(r);
			
			r = new ELPRule();
			r.addHead( new ELPAtom("ok","0"));
			r.addBody( new ELPAtom("ok",""+j));
			ret.addRule(r);
		}
		
		ELPRule r = new ELPRule();
		r.addBody( new NotLiteral( new ELPAtom("ok","0")));
		ret.addRule(r);
		
		r = new ELPRule();
		r.addHead( new ELPAtom("eq", ""+(imax+1)));
		ret.addRule(r);
		
		return ret;
	}
	
	/**
	 * this method returns the rej-Literal, if
	 * it is in the head of a rule
	 * 
	 * @param r rule to test
	 * @return reject literal or null
	 */
	protected ELPAtom isRejRule(ELPRule r) {
		for (int i = 0; i < r.nHead(); i++) {
			ELPAtom atm = r.getHead(i).getAtom();
			if (atm.getPredicate().startsWith("rej"))
				return atm;
		}
		return null;
	}

	@Override
	public Map<String, List<ELPLiteral>> getRejectSet() {
		// not supported, rejected literals are
		// processed in selectAnswerSet
		return new HashMap<String, List<ELPLiteral>>();
	}

	@Override
	public void setSavePredicates(Set<String> savePreds) {
		// not supported
	}

}
