package edu.udo.cs.ie.cowbots.logic;

import jason.asSyntax.*;

import java.io.*;
import java.util.*;

import edu.udo.cs.ie.cowbots.bdi.brf.Parser;
import edu.udo.cs.ie.cowbots.logic.parser.elp.ELPParser;

/**
 * this class models an extended logic program. it
 * consists of a set of rules (and fact and constraints).
 * 
 * @author Thomas
 *
 */
public class ELP {
	
	ArrayList<ELPRule>	rules = new ArrayList<ELPRule>();;
	
	public ELP(Literal l) {
		//Parser parser = new Parser();
		rules.addAll(Parser.parseJLiteralToRules(l)) ;
		
	}
	
	public ELP() {
	}
	
	public void addRule(ELPRule r) {
		rules.add(r);
	}
	
	public String toString() {
		StringBuilder sb = new StringBuilder();
		
		for(ELPRule r : rules) {
			sb.append(r.toString()+"\n");
		}
		
		return sb.toString();
	}
	
	/**
	 * number of rules in elp
	 * @return
	 */
	public int nRules() {
		return rules.size();
	}
	
	/**
	 * get rule i
	 * @param i
	 * @return	rule with name (index) i
	 */
	public ELPRule getRule(int i) {
		return rules.get(i);
	}
	
	/**
	 * returns all elp rules in this program
	 *  
	 * @return	arraylist of elp rules
	 */
	public ArrayList<ELPRule> getRules() {
		return this.rules;
	}
	
	
	/**
	 * return all literals that are facts (even disjunctive)
	 * @return
	 */
	public List<ELPLiteral> getFacts() {
		ArrayList<ELPLiteral> ret = new ArrayList<ELPLiteral>();
		for (ELPRule r : rules) {
			if (!r.isFact())
				continue;
			
			int iI = r.nHead();
			for (int i = 0; i < iI; i++) {
				ret.add(r.getHead(i));
			}
		}
		return ret;
	}
	
	/**
	 * store elp on disc
	 * 
	 * @param file	output filename (like foo.elp or bar.txt)
	 */
	public void saveAs(String filename) {
		try {
			BufferedWriter w = new BufferedWriter(new FileWriter(filename));
			for (ELPRule r : rules) {
				w.write(r.toString());
				w.newLine();
			}
			w.flush();
			w.close();
		} catch (Exception e) {
			
		}
	}
	
	/**
	 * loads an elp from a file ressource
	 * 
	 * @param filename file to read elp from
	 */
	public static ELP loadFrom(String filename) {
		ELP ret = new ELP();
		
		try {
			
			return ELPParser.parse(filename);
			
		} catch (Exception e) {
			System.out.println("parser error while parsing "+filename+"!");
			e.printStackTrace();
		}
		
		return ret;
	}
	
	/**
	 * merges this and another elp and return a new elp
	 * 
	 * @param other	second elp to merge with
	 * @return elp with all rules from this and other
	 */
	public ELP merge(ELP other) {
		ELP ret = new ELP();		
		
		ret.rules.addAll(this.rules);
		if (other != null)
			ret.rules.addAll(other.rules);
		
		return ret;
	}
	
	
	/**
	 * adds a raw, unchecked string as the head of a new
	 * rule to the elp. this method can be used to pass
	 * arbitrary statements to the program. it also
	 * removes the last '.' from the input string.
	 * anyway, use with care..
	 * 
	 * @param s		text string added to program
	 */
	public void addRaw(String s) {
		if (s.endsWith("."))
			s = s.substring(0,s.length()-1);
		ELPRule r = new ELPRule(1,0);
		r.addHead( new ELPAtom(s));
		this.addRule(r);
	}
	
	
	/**
	 * returns the alphabet of this program (set
	 * of all functor symbols)
	 * 
	 * @return	alphabet of program
	 */
	public Alphabet	getAlphabet() {
		return new Alphabet(this);
	}
	
	
	/**
	 * splits the given program into rules and facts.
	 * note that disjunctive facts are treated as rules.
	 *  
	 * @param literals	output container for all literals
	 * @param rules		output container for all rules
	 */
	public void split(Collection<ELPLiteral> literals, Collection<ELPRule> rules) {
		for(ELPRule r : this.rules) {
			if (r.nHead()==1 && r.nBody()==0)
				literals.add(r.getHead(0));
			else
				rules.add(r);
		}
	}
	
	/**
	 * this methods adds all rules from the given
	 * program to this program
	 * @param program2	program to add
	 */
	public void add(ELP program2) {
		if (program2 == null)
			return;
		
		this.rules.addAll(program2.getRules());
	}
}
