package edu.udo.cs.ie.cowbots.util.cluster;

import java.awt.Point;
import java.util.*;
import java.util.Map.Entry;
import edu.udo.cs.ie.cowbots.internalActions.scout.astar;
import edu.udo.cs.ie.cowbots.logic.*;
import edu.udo.cs.ie.cowbots.util.DistanceMap;
import edu.udo.cs.ie.cowbots.util.GridMap;


/**
 * Utility class for calculating cow clusters identified
 * by one intelligent selected cow for driving.
 * 
 * @author dhoelzgen
 * @author Thomas, Michael (new cluster method)
 *
 */
public class ClusterCalculator implements IClusterCalculator {

	static final int defaultCulsterDistance = 4;
	private int clusterDistance = 4;

	private astar pathfinder;

	private HashMap<String, Integer> cowToCluster;
	private HashMap<String, Point> cowPositions;
	

	/* (non-Javadoc)
	 * @see edu.udo.cs.ie.cowbots.util.IClusterCalculator#getCowClusters(edu.udo.cs.ie.cowbots.util.GridMap, int, int)
	 */
	public List<CowCluster> getCowClusters(GridMap map, DistanceMap dm, int cDist, int maxCowTime) {
		
		clusterDistance = cDist;
		int nextCluster = 0;

		pathfinder = new astar();
		pathfinder.cluster_agentwall = true;

		cowToCluster = new HashMap<String, Integer>();
		cowPositions = new HashMap<String, Point>();

		for (int i = 0; i < map.width(); i++) {
			for (int j = 0; j < map.height(); j++) {
				if (map.getContent(i, j).getAtom().getTerm(0).contains("cow")) {
					String cowName = map.getContent(i, j).getAtom().getTerm(0);
					
					// if the cow is in a corner, ignore it
					if (map.isCorner(i, j))
						continue;
					
					cowPositions.put(cowName, new Point(i, j));
					int foundCluster = getCowWithClusterInNeighbourhood(i, j, map);

					if (foundCluster == -1) {
						cowToCluster.put(cowName, nextCluster);
						nextCluster++;
					} else {
						cowToCluster.put(cowName, foundCluster);
					}

				}
			}
		}

		// TODO: Added for experimental testing, could be changed to provide more information
		// TODO: No matter how, this should be coded in another way...
		HashMap<Integer, List<String>> result = new HashMap<Integer, List<String>>();

		for (Entry<String, Integer> e : cowToCluster.entrySet()) {
			if (!result.containsKey(e.getValue())) {
				List<String> newEntry = new ArrayList<String>();
				newEntry.add(e.getKey());
				result.put(e.getValue(), newEntry);
			} else {
				List<String> currentList = result.get(e.getValue());
				currentList.add(e.getKey());
				result.put(e.getValue(), currentList);
			}
		}
		
		List<CowCluster> cluList = new LinkedList<CowCluster>();
		for (List<String> cowList : result.values()) {
			Iterator<String> cowIter = cowList.iterator();
			if (cowIter.hasNext()) {
				String cname = cowIter.next();
				Point cp = cowPositions.get(cname);
				CowCluster cc = new CowCluster( new CowPosition(cp.x,cp.y,cname) );
				while (cowIter.hasNext()) {
					cname = cowIter.next();
					cp = cowPositions.get(cname);
					cc.tryAdd( new CowPosition(cp.x, cp.y, cname), 200);
				}
				cluList.add(cc);
			}
		}

		return cluList;
	}

	/**
	 * Uses a* to find a way to another cow wich is already assigned to a cluster.
	 * Treats agents and their surrounding as obstacles, so they can split clusters.
	 * 
	 * @param x
	 * @param y
	 * @param map
	 * @return
	 */
	protected int getCowWithClusterInNeighbourhood(int x, int y, GridMap map) {
		
		try {
			for (Entry<String, Integer> e : cowToCluster.entrySet()) {
				Point potentialTargetCow = cowPositions.get(e.getKey());
				int tX = potentialTargetCow.x;
				int tY = potentialTargetCow.y;
				if (map.isCorner(tX, tY))
					return -1;
 			
				int cDist = pathfinder.getPath(map, x, y, tX, tY, false).length;
				if (cDist <= clusterDistance && cDist > 0)
					return e.getValue();
			}
		} catch (Exception ex) {}

		return -1;
	}
	
	
	/**
	 * a very simple cluster method. for every cow, it creates a
	 * new cluster or puts the cow in the first cluster while
	 * keeping a maximal cluster width (or height).
	 * 
	 * @param map
	 * @param worldtime
	 * @return
	 */
	public HashMap<Integer, List<String>> newClusterMethod(GridMap map, int worldtime) {
		
		// list of cow position
		List<ELPLiteral> cowPositions = map.getMoversByType("cow");
		
		// List of all clusters found
		List<CowCluster> cowClusters = new LinkedList<CowCluster>();
		
		// map of all unclustered cows
		List<CowPosition> unclusteredCows = new LinkedList<CowPosition>();
				
		// max. allowed age for a cow in a cluster
		int timelimit = 0;
		
		// max. distance between cows in a cluster
		int clusterDistance = 4;
		
		// initialization - skip cows that are to old
		for (ELPLiteral l : cowPositions) {
			ELPAtom atm = l.getAtom();
			if (timelimit > 0)
				if (Integer.parseInt(atm.getTerm(3)) < (worldtime-timelimit))
					continue;
			CowPosition cp = new CowPosition(atm);
			unclusteredCows.add(cp);
		}
				
		
		// loop over all cows
		Iterator<ELPLiteral> cowIter = cowPositions.iterator();
		while( cowIter.hasNext() ) {
			
			// remove cow which starts a cluster
			ELPLiteral currentCow = cowIter.next();
			CowPosition cp = new CowPosition(currentCow);
			 				
			// try finding a cow cluster
			boolean clusterFound = false;
			for (CowCluster cc : cowClusters) {
				if (cc.tryAdd(cp, clusterDistance)) {
					clusterFound=true;
					break;
				}
			}
			// no cluster found, create a new one
			if (!clusterFound)
				cowClusters.add( new CowCluster(cp));
		}
		
		// prepare data to match getCowClusters method
		HashMap<Integer, List<String>> ret = new HashMap<Integer, List<String>>();
		int clusterCounter = 0;
		for (CowCluster cc : cowClusters) {
			List<String> cowNames = new LinkedList<String>();
			for (CowPosition cp : cc.cows)
				cowNames.add(cp.name);
			ret.put(clusterCounter, cowNames);
			++clusterCounter;
		}
		return ret;
	}
	
}
