package edu.udo.cs.ie.cowbots.util.cluster;

import java.util.LinkedList;
import java.util.List;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Point;

import edu.udo.cs.ie.cowbots.util.PathTransformer;

/**
 * auxiliary class to build cow clusters. a cow cluster
 * is a set of 1 to n cows that stay within a rectangular
 * area of cells of a specific width.
 * 
 * @author Thomas
 *
 */
public class CowCluster {
	
	// list of cows assigned to this cluster
	public List<CowPosition>	cows;
	
	// cluster boundaries
	public int minX, minY, maxX, maxY;

	// acceptedCluster is a hint for the debug output
	// if the cluster is accepted by an agent or
	// not. every accepted cluster is part of
	// the desire generation
	public boolean	acceptedCluster;
	// targetCluster is true if this cluster should
	// be driven into the corral
	public boolean	targetCluster;
	
	// these variables hold an associated path to
	// drive the cluster generated by a* or similiar
	// pathfinding methods
	public String[]	path;
	public int		pathStartX, pathStartY;
	
	// best positions around the cluster to drive the cow as intended
	protected List<Point>	drivePoints;
	
	/**
	 * default constructor for a cluster. a cluster always
	 * starts with at least one cow.
	 * 
	 * @param cp
	 */
	public CowCluster(CowPosition cp) {
		cows = new LinkedList<CowPosition>();
		cows.add(cp);
		minX = cp.x;
		maxX = cp.x;
		minY = cp.y;
		maxY = cp.y;
		acceptedCluster = false;
		targetCluster = false;
		path = null;
		drivePoints = null;
	}
	
	public boolean tryAdd(CowPosition cp, int maxSize) {
		int tmpMinX = minX;
		int tmpMinY = minY;
		int tmpMaxX = maxX;
		int tmpMaxY = maxY;
		
		if (cp.x < tmpMinX)
			tmpMinX = cp.x;
		if (cp.x > tmpMaxX)
			tmpMaxX = cp.x;
		if (cp.y < tmpMinY)
			tmpMinY = cp.y;
		if (cp.y > tmpMaxY)
			tmpMaxY = cp.y;
		
		int w = tmpMaxX-tmpMinX+1;
		int h = tmpMaxY-tmpMinY+1;
		if ((w <= maxSize) && (h<=maxSize)) {
			this.cows.add(cp);
			
			minX = tmpMinX;
			minY = tmpMinY;
			maxX = tmpMaxX;
			maxY = tmpMaxY;
			
			return true;
		} else
			return false;
	}
	
	/**
	 * this method visualizes the cluster on a graphics2d
	 * device. used for debug output in the lib_cowbots
	 * framework.
	 * @param g2
	 */
	public void draw(Graphics2D g2, int cellX, int cellY) {
		if (!acceptedCluster)
			g2.setColor( Color.RED);
		else
			g2.setColor( Color.GREEN);
		g2.drawRect( minX * cellX , minY * cellY, (maxX-minX+1)*cellX, (maxY-minY+1)*cellY);
		
		int cellXo2 = cellX/2;
		int cellYo2 = cellY/2;
		
		// drawpath to corral
		if (path != null) {
			int x0 = pathStartX;
			int y0 = pathStartY;
			for (String dir : path) {
				int x1 = PathTransformer.getDirectionX(dir) + x0;
				int y1 = PathTransformer.getDirectionY(dir) + y0;
				g2.drawLine(x0*cellX+cellXo2,y0*cellY+cellYo2,x1*cellX+cellXo2,y1*cellY+cellYo2);
				x0 = x1;
				y0 = y1;
			}
		}
		
		if (targetCluster) {
			java.util.List<HullPoint> ppList = calculatePinchPoints(3);
			for (HullPoint p : ppList) {

				int col = 100+(int) (p.sim*75.0);
				g2.setColor( new Color(col,col,0));
				if (!p.marked)
					g2.drawRect( p.x*cellX+2, p.y*cellY+2, cellX-4, cellY-4);
			}
		}
	}
	

	
	/**
	 * this method returns the x-component of
	 * the center of a corral cluster based 
	 * on the cluster's bounding rectangle center.
	 * @return	x-coordinate of cluster center
	 */
	public int centerX() {
		return minX + (maxX-minX)/2;
	}
	
	/**
	 * this method returns the y-component of
	 * the center of a corral cluster based 
	 * on the cluster's bounding rectangle center.
	 * @return	y-coordinate of cluster center
	 */
	public int centerY() {
		return minY + (maxY-minY)/2;
	}
	
	/**
	 * returns the maximum elongation along
	 * the x-axis of the cluster
	 * 
	 * @return cluster's horizontal elongation
	 */
	public int lengthX() {
		return maxX-minX+1;
	}
	
	
	/**
	 * returns the maximum elongation along
	 * the y-axis of the cluster
	 * 
	 * @return cluster's vertical elongation
	 */
	public int lengthY() {
		return maxY-minY+1;
	}
	
	/**
	 * this method calculates positions for
	 * agents to drive the cluster in the
	 * desired direction. the desired direction
	 * is taken from the path associated with
	 * the cluster.  
	 * 
	 * @param impact approximated effective influence radius of an agent against a cow
	 */
	public java.util.List<HullPoint> calculatePinchPoints(int impact) {
		if (this.path == null)
			return null;
		if (this.path.length == 0)
			return null;
		
		LinkedList<HullPoint> ppoints = new LinkedList<HullPoint>();
		
		int driveX = PathTransformer.getDirectionX(path[0]) + pathStartX;
		int driveY = PathTransformer.getDirectionY(path[0]) + pathStartY;
		
		// hullpoints represent potential agent positions around a cluster
		int xLen = lengthX();
		int yLen = lengthY();		
		HullPoint hullpoints[] = new HullPoint[xLen*2+yLen*2+4];		
		xLen += 2;
		yLen += 2;
		
		// fill hullpoints with initial values, which is the
		// cosine between the drive vector and a center to hull-point
		// vector. values near one are suspected to be next clow escape
		// points, values near minus one are good positions in the
		// clusters back. 
		
		int hullIndex = 0;
		int cX = centerX();		
		int cY = centerY();
		
		// create & link hull points
		hullpoints[hullIndex++] = new HullPoint(minX-1,minY-1);
		
		for (int i = minX; i <= maxX+1; i++) {
			hullpoints[hullIndex++] = new HullPoint(i,minY-1);
			hullpoints[hullIndex-2].right = hullpoints[hullIndex-1];
			hullpoints[hullIndex-1].left = hullpoints[hullIndex-2];
		}
		
		for (int i = minY; i <= maxY; i++) {
			hullpoints[hullIndex++] = new HullPoint(maxX+1,i);
			hullpoints[hullIndex-2].right = hullpoints[hullIndex-1];
			hullpoints[hullIndex-1].left = hullpoints[hullIndex-2];
		}
		
		for (int i = maxX+1; i >= minX-1; i--) {
			hullpoints[hullIndex++] = new HullPoint(i,maxY+1);
			hullpoints[hullIndex-2].right = hullpoints[hullIndex-1];
			hullpoints[hullIndex-1].left = hullpoints[hullIndex-2];
		}
		
		for (int i = maxY; i >= minY; i--) {
			hullpoints[hullIndex++] = new HullPoint(minX-1,i);
			hullpoints[hullIndex-2].right = hullpoints[hullIndex-1];
			hullpoints[hullIndex-1].left = hullpoints[hullIndex-2];
		}

		hullpoints[0].left = hullpoints[hullIndex-1];
		hullpoints[hullIndex-1].right = hullpoints[0];
		
		
		// compute similarities
		HullPoint escapeStart = null;
		HullPoint impactStart = null;
		
		HullPoint hp = hullpoints[0];
		hp.computeSimilarity(driveX, driveY, cX, cY);
		escapeStart = hp;
		impactStart = hp;
		ppoints.add( hp );		
		hp.marked = true;
		
		while (!hp.right.marked) {
			hp = hp.right;
			hp.computeSimilarity(driveX, driveY, cX, cY);
			if (hp.sim > escapeStart.sim)
				escapeStart = hp;
			if (hp.sim < impactStart.sim)
				impactStart = hp;
			
			ppoints.add( hp);
			
			hp.marked = true;
		}
		
		// unmark the adjacent cells of the escapeStart
		// allowed escape range is equal to the impact
		// of agents pinching the cluster
		int escapeZone = impact;
		// fix escape zone for small hulls
		if (impact*2 >= hullpoints.length/2)
			escapeZone /= 2;
		
		escapeStart.marked = false;
		HullPoint hullIter = escapeStart.right;
		for (int i = 0; i < escapeZone; i++) {
			hullIter.marked = false;
			hullIter = hullIter.right;
		}
		
		hullIter = escapeStart.left;
		for (int i = 0; i < escapeZone; i++) {
			hullIter.marked = false;
			hullIter = hullIter.left;
		}
		
		hullIter = impactStart;
		hullIter.marked = false;
		hullIter = hullIter.right;
		int iCount = impact;
		while ( hullIter.marked == true ) {
			--iCount;
			if (iCount <= 0) {
				hullIter.marked = false;				
				iCount = impact;
			}
			hullIter = hullIter.right;
		}
		
		hullIter = impactStart;
		hullIter.marked = false;
		hullIter = hullIter.left;
		iCount = impact;
		while ( hullIter.marked == true ) {
			--iCount;
			if (iCount <= 0) {
				hullIter.marked = false;
				iCount = impact;
			}
			hullIter = hullIter.left;
		}
				
		return ppoints;
	}
	
	
	/**
	 * auxiliary class that holds a hull point and
	 * it's adjacent hull points
	 * 
	 * @author Thomas
	 *
	 */
	class HullPoint {
		public int x,y;
		public HullPoint left, right;
		public boolean marked;
		public double sim;
		
		public HullPoint(int x, int y) {
			this.x = x;
			this.y = y;
			this.marked = false;
			this.left = null;
			this.right = null;
		}

		/**
		 * this method computes the similarity of a hull point
		 * compared to a drive vector. similarity computation
		 * based on the scalar product of two vectors.
		 * @param cX
		 * @param cY
		 */
		public void computeSimilarity(int dX, int dY,int cX, int cY) {
			double x0 = cX-dX;
			double y0 = cY-dY;
			double x1 = cX-x;
			double y1 = cY-y;
			double numerator = x0*x1+y0*y1;
			double denumerator = Math.sqrt(x0*x0+y0*y0)*Math.sqrt(x1*x1+y1*y1);
			double cosine = 1.0;
			if (Math.abs(denumerator) <= 0.0001) {
				// do not divide by near-zero
			} else {
				cosine = numerator/denumerator;
			}
			this.sim = cosine;
		}
				
	}
	
	/**
	 * returns a list of probably optimal positions
	 * for driving the cow cluster in the desired 
	 * direction.
	 * 
	 * @return list of agent positions
	 */
	public List<Point> getAgPositions() {
		if (this.drivePoints != null)
			return this.drivePoints;
		
		drivePoints = new LinkedList<Point>();
		HullPoint lowestSim = null;
		
		for (HullPoint hp : calculatePinchPoints(3) ) {
			if (lowestSim == null)
				lowestSim = hp;
			else if (hp.sim < lowestSim.sim)
				lowestSim = hp;
			
			drivePoints.add( new Point( hp.x, hp.y ));
		}
		
		drivePoints.add(0, new Point(lowestSim.x, lowestSim.y));
		return drivePoints;
	}
}